
def to_dict(obj, class_key=None):
    if isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            data[k] = to_dict(v, class_key)
        return data
    elif hasattr(obj, "_ast"):
        return to_dict(obj._ast())
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [to_dict(v, class_key) for v in obj]
    elif hasattr(obj, "__dict__"):
        data = dict([(key, to_dict(value, class_key))
            for key, value in obj.__dict__.items()
            if not callable(value) and not key.startswith('_')])
        if class_key is not None and hasattr(obj, "__class__"):
            data[class_key] = obj.__class__.__name__
        return data
    else:
        return obj





# TODO: Create all of the EOSIO types using numpy and either colander/marshmallow

# EOSIO Types

# block_time_stamp


# Names
# see eos/libraries/chain/include/eosio/chain/name.hpp?L#21

# Name : uint64 bit encoded name type or string
# AccountName : Name Type
# ScopeName : Name Type
# PermissionName : Name Type
# ActionName : Name Type
# TableName : Name Type

# Native Structs

# Authority Struct : {'threshold': uint32, 'keys': [KeyWeight], 'accounts': [AccountWeight], 'waits': [WaitWeight]}
# KeyWeight Struct : {'key': PublicKey, 'weight': uint16}
# WaitWeight Struct : {'wait_sec': uint32, 'weight': uint16}
# Asset Struct : {'amt': int64, 'symbol': Symbol}
# Symbol Struct : {'precision': ubyte, 'name': string} - uint64 encoded precision and name

# ABI.ABI Structs

# ABI Struct : {'version': string, 'types': [ABIType], 'structs': [StructType],
#                   'actions': [ActionType], 'tables': [TableType], 'clauses': [ClauseType],
#                   'error_messages': [ABIErrorMessageType], 'abi_extensions': [ExtensionsType]}

# ABIType : {'new_type_name': string, 'type': type}

# StructType : {'name': string, 'base': string, 'fields': [FieldType]}

# FieldType : {'name': string, 'type': string}

# ActionType : {'name': ActionName, 'type': string, 'ricardian_contract': string}

# TableDef : {'name': TableName, 'index_type': string, 'key_names': [string], 'key_types': [string], 'type': string}

# ClauseType : {'id': string, 'body': string}

# ABIErrorMessageType : {'error_code': uint64, 'error_msg': string}





#    _                _                      _ __    _  _
#   | |_     ___     | |     ___     ___    | '_ \  | || |
#   |  _|   / -_)    | |    / _ \   (_-<    | .__/   \_, |
#   _\__|   \___|   _|_|_   \___/   /__/_   |_|__   _|__/
# _|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_| """"|
# "`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'


import logging

logger = logging.getLogger()
# TODO: setup a method for switching the code from development to production
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)

logger.addHandler(ch)




"""
eosPython exceptions
"""
from requests.exceptions import RequestException

__all__ = ['APIResponseException', 'WalletAPIException', 'ChainAPIException', 'APINotConfigured', 'ModelException',
           'PermissionDoesNotExistException', 'AccountDoesNotExistException', 'AccountABIDoesNotExistException',
           'ActionDoesNotExistException', 'AccountAlreadyExistsException']

# TODO: write handlers, that are used by the API(s) to raise verbose exceptions based on
# information returned by the RPC API


def wallet_exception_raiser():
    pass


def chain_exception_raiser():
    pass


################
# API EXCEPTIONS
################


class APIResponseException(RequestException):
    """API Exception thrown a when response status code is 500 or greater"""

    def __str__(self):
        if self.response is not None:
            return str(self.response.json())
        return "Whoops! An error occurred..."


class APINotConfigured(IOError):
    """This API isn't configured correctly"""


################
# WALLET EXCEPTIONS
################


class WalletAPIException(APIResponseException):
    """Wallet API Exception thrown when a response status code is 500 or greater"""


class WalletAlreadyExistsException(WalletAPIException):
    pass


class WalletDoesNotExistException(WalletAPIException):
    pass


class WalletLockedException(WalletAPIException):
    pass


class WalletUnlockedException(WalletAPIException):
    pass


class WalletMissingPubKeyException(WalletAPIException):
    pass


class WalletInvalidPasswordException(WalletAPIException):
    pass


class WalletNotAvailableException(WalletAPIException):
    pass


class KeyExistsException(WalletAPIException):
    pass


class UnsupportedKeyTypeException(WalletAPIException):
    pass


class InvalidLockTimeoutException(WalletAPIException):
    pass


class SecureEnclaveExceptions(WalletAPIException):
    pass


################
# CHAIN EXCEPTIONS
################


class ChainAPIException(APIResponseException):
    """Chain API Exception thrown when a response status code is 500 or greater"""


# Authorization Exceptions


class AuthorizationException(ChainAPIException):
    pass


class DuplicateSignatureException(ChainAPIException):
    pass


class UnsatisfiedAuthException(ChainAPIException):
    pass


class MissingAuthException(ChainAPIException):
    pass


class IrrelevantAuthException(ChainAPIException):
    pass


class InsufficientDelayException(ChainAPIException):
    pass


class InvalidPermissionException(ChainAPIException):
    pass


class InvalidParentPermission(ChainAPIException):
    pass


# Resource Exceptions

class ResourceExhaustedException(ChainAPIException):
    pass


class RamUsageExceededException(ChainAPIException):
    pass


class TXNetUsageExceededException(ChainAPIException):
    pass


class BlockNetUsageExceededException(ChainAPIException):
    pass


class TXCPUUsageExceededException(ChainAPIException):
    pass


class BlockCPUUsageExceededException(ChainAPIException):
    pass


class DeadlineException(ChainAPIException):
    pass


class GreylistNetUsageExceeded(ChainAPIException):
    pass


# Transaction Exceptions


class TransactionException(ChainAPIException):
    pass


class DecompressionException(TransactionException):
    pass


class NoActionExceptions(TransactionException):
    pass


class NoAuthorizations(TransactionException):
    pass


class CFAAuthorizationException(TransactionException):
    pass


class ExpiredTXException(TransactionException):
    pass


class InvalidRefBlockException(TransactionException):
    pass


class DuplicateTXException(TransactionException):
    pass


class DuplicateDeferredTXException(TransactionException):
    pass


class CFAInsideGeneratedTXException(TransactionException):
    pass


class TXNotFoundException(TransactionException):
    pass


class TooManyTXException(TransactionException):
    pass


class TXTooBigException(TransactionException):
    pass


class UnknownCompression(TransactionException):
    pass


# Type Exceptions

class TypeException(ChainAPIException):
    pass


class NameTypeException(TypeException):
    pass


class PublicKeyTypeException(TypeException):
    pass


class PrivateKeyTypeException(TypeException):
    pass


class AuthorityTypeException(TypeException):
    pass


class ActionTypeException(TypeException):
    pass


class TransactionTypeException(TypeException):
    pass


class ABITypeException(TypeException):
    pass


class BlockIDTypeException(TypeException):
    pass


class TransactionIDTypeException(TypeException):
    pass


class PackedTransactionTypeException(TypeException):
    pass


class AssetTypeException(TypeException):
    pass


class ChainIDTypeException(TypeException):
    pass


class FixedKeyTypeException(TypeException):
    pass


class SymbolTypeException(TypeException):
    pass


################
# MODEL EXCEPTIONS
################


class ModelException(IOError):
    """Model Exception thrown when a model's validations fail"""


class PermissionDoesNotExistException(ModelException):
    """Permission DNE Exception thrown when a specified permission does not exist on a specified account"""


class AccountDoesNotExistException(ModelException):
    """Account DNE Exception thrown when a specified account_name does not exist"""


class AccountAlreadyExistsException(ModelException):
    """Account DNE Exception thrown when a specified account_name does not exist"""


class AccountABIDoesNotExistException(ModelException):
    """ABI DNE Exception thrown when an account does not have an ABI"""


class ActionDoesNotExistException(ModelException):
    """Action DNE Exception thrown when account ABI does not contain action_name"""



from datetime import datetime
from datetime import timedelta
import logging

__all__ = ['Transaction', 'AccountName', 'Action', 'Permission']

# TODO: create schemas using either marshmallow or colander


class ModelBase(object):
    """Base class for all eosPython models"""


class Asset(ModelBase):
    pass


class Transaction(ModelBase):
    logger = logging.getLogger(__name__)

    def __init__(self, *initial_data, **kwargs):
        """Transaction - The Python Object representation of an EOSIO Transaction"""
        self.expiration = (datetime.utcnow() + timedelta(minutes=2)).isoformat()
        self.ref_block_num = 0
        self.ref_block_prefix = 0
        self.max_net_usage_words = 0
        self.max_cpu_usage_ms = 0
        self.delay_sec = 0
        self.context_free_actions = []
        self.context_free_data = []
        self.actions = []
        self.signatures = []

        for dictionary in initial_data:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def set_ref_data(self, chain_info, lib_info):
        if chain_info is not None:
            self.ref_block_num = chain_info['last_irreversible_block_num']
        if lib_info is not None:
            self.ref_block_prefix = lib_info['ref_block_prefix']

    def add_action(self, action):
        if not self.actions:
            self.actions = []
        self.actions.append(action)


class AccountName(ModelBase):

    def __init__(self, name):
        """Python object representation of an EOSIO account_name"""
        assert isinstance(name, str), "Argument name must be type string"
        self.account_name = name


# TODO: Create an authority object that can be used for updating account permissions


class Permission(ModelBase):

    def __init__(self, actor, permission):
        """Authority is the account_name and permission name used to authorize an action"""
        assert isinstance(actor, str), "Argument `actor` must be type String"
        assert isinstance(permission, str), "Argument `permission` must be type String"
        self.actor = actor
        self.permission = permission


class Action(ModelBase):
    logger = logging.getLogger(__name__)

    def __init__(self, account, action_name, args, authorizations=[]):
        """Action is used in pushing transactions to the RPC API"""
        if isinstance(authorizations, Permission):
            authorizations = [authorizations]
        elif not isinstance(authorizations, list):
            raise AssertionError('authorizations must be a list of Authority or a single Authority')
        assert isinstance(account, str), "Argument account must be type String"
        assert isinstance(action_name, str), "Argument action_name must be type String"
        self.account = account  # NOTE: code, is the account_name the contract is set on.
        self.code = account
        self.name = action_name
        self.action = action_name
        self.authorization = []  # NOTE: Authorization is the permission_level used for the action
        self.args = args
        self.data = None  # NOTE: Data is the binargs received from abi_json_to_bin RPC

        for auth in authorizations:
            self.add_authorization(auth)

    def add_authorization(self, authority):
        self.authorization.append(authority)




#!/usr/bin/env python
import os
import sys
from codecs import open
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

if sys.argv[-1] == 'publish':
    os.system('python3 setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

packages = ['telospy']

requires = [
    'requests>=2.19.1'
]

about = {}
with open(os.path.join(here, 'telospy', '__version__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()

# with open('HISTORY.rst', 'r', 'utf-8') as f:
#     history = f.read()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=packages,
    package_data={'': ['LICENSE', 'NOTICE'], 'telospy': ['*.pem']},
    package_dir={'telospy': 'telospy'},
    include_package_data=True,
    python_requires=">=2.6, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    install_requires=requires,
    license=about['__license__'],
    zip_safe=False
)



import logging
import requests
from urllib import parse
from .models import *
from .exceptions import *
from .util import to_dict

__all__ = ['WalletAPI', 'ChainAPI', 'API']


class API:

    def __init__(self, url, wallet_url='http://127.0.0.1:8999', version='v1'):
        self._wallet_api = None
        self._chain_api = None
        self.logger = logging.getLogger(__name__)

        assert isinstance(url, str), "`url` is not a string"
        self.logger.debug('Configuring chain_api instance')
        self._chain_api = ChainAPI(url, version)

        assert isinstance(wallet_url, str), "`wallet_url` is not a string"
        self.logger.debug('Configuring wallet_api instance')
        self._wallet_api = WalletAPI(wallet_url, version)

    @property
    def chain(self):
        if not self._chain_api:
            self.logger.error(msg='_chain_api is None')
            raise APINotConfigured('Chain API has not yet been configured. api.configure_chain_api(url)')
        self.logger.debug('returning ChainAPI instance')
        return self._chain_api

    @property
    def wallet(self):
        if not self._wallet_api:
            self.logger.error(msg='_wallet_api is None')
            raise APINotConfigured('Wallet API has not yet been configured. api.configure_wallet_api()')
        self.logger.debug('returning WalletAPI instance')
        return self._wallet_api

    def get_chain_lib_info(self):
        """Gets required chain information for trx"""
        chain_info = self.chain.get_info()
        lib_info = self.chain.get_block(chain_info['last_irreversible_block_num'])
        return chain_info, lib_info

    def get_account_info(self, account_name):
        try:
            return self.chain.get_account(account_name)
        except ChainAPIException:  # NOTE: This exception is too specific
            raise AccountDoesNotExistException('The account does not exist on this chain')

    def does_action_exist(self, account_name, action_name, account_info={}):
        assert isinstance(account_name, str), "`account_name` must be a String"
        if isinstance(account_info, dict) and 'account_name' in account_info:
            assert account_name == account_info['account_name'], \
                'the `account_information given does not match the provided account_name`'
            account_name = account_info['account_name']
        abi_info = self.chain.get_abi(account_name)
        if 'abi' not in abi_info:
            raise AccountABIDoesNotExistException('account_name `%s` does not have an ABI' % account_name)

        if 'actions' in abi_info['abi']:
            for action in abi_info['abi']['actions']:
                if action['name'].lower() == action_name.lower():
                    return True
        return False

    def does_permission_exist(self, account_info, permission_name):
        if 'permissions' in account_info:
            for permission in account_info['permissions']:
                if permission['perm_name'].lower() == permission_name.lower():
                    return True
        return False

    def get_bin_data(self, action):
        return self.chain.abi_json_to_bin(action)['binargs']

    ####################
    # Basic Chain Commands
    ####################

    # TODO: refactor to be more verbose. Where owner and active can be either keys or account permissions
    # TODO: refactor this use new Authority model instead of dictionaries
    def create_account(self, creator, account_name, owner, active='', stake_net='1000.0000 TLOS',
                       stake_cpu='1000.0000 TLOS',
                       ram_kb=1024, permission='active', transfer=True):
        try:
            self.get_account_info(account_name)
            raise AccountAlreadyExistsException('The account already exists on this chain')
        except AccountDoesNotExistException:
            pass

        owner_auth = {"threshold": "1", "keys": [{"key": owner, "weight": 1}], 'accounts': [], "waits": []}
        if active is '':
            active_auth = owner_auth
        else:
            active_auth = {"threshold": "1", "keys": [{"key": owner, "weight": 1}], 'accounts': [], "waits": []}

        create_action = Action('eosio', 'newaccount',
                               {'creator': creator, 'name': account_name, 'owner': owner_auth, 'active': active_auth},
                               Permission(creator, permission))
        buy_ram_action = Action('eosio', 'buyrambytes',
                                {'payer': creator, 'receiver': account_name, 'bytes': ram_kb * 1024},
                                Permission(creator, permission))
        delegate_bw_action = Action('eosio', 'delegatebw',
                                    {'from': creator, 'receiver': account_name, 'stake_net_quantity': stake_net,
                                     'stake_cpu_quantity': stake_cpu, 'transfer': transfer},
                                    Permission(creator, permission))

        return self.send_transaction(actions=[create_action, buy_ram_action, delegate_bw_action])

    def set_contract(self):
        pass

    def transfer(self, sender, recipient, amount, memo):
        pass

    def updateauth(self):
        pass

    def send_transaction(self, actions):
        """Send a transaction with a single action in it."""
        if isinstance(actions, Action):
            actions = [actions]
        elif not isinstance(actions, list):
            raise AssertionError('actions must be a list of Actions or a single Action')

        chain_info, lib_info = self.get_chain_lib_info()

        for action in actions:
            for auth in action.authorization:
                auth_info = self.get_account_info(auth.actor)
                if not self.does_permission_exist(auth_info, auth.permission):
                    raise PermissionDoesNotExistException('account_name `%s` does not have permission `%s`'
                                                          % (auth.actor, auth.permission))
            code_info = self.get_account_info(action.account)  # NOTE: this method will raise an exception if it fails

            if not self.does_action_exist(action.code, action.name, code_info):
                raise ActionDoesNotExistException('`%s` ABI does not contain an action named `%s`'
                                                  % (action.code, action.name))
            action.data = self.get_bin_data(to_dict(action))
            self.logger.debug('action bin data: {}'.format(action.data))

        trx = Transaction(actions=actions)
        trx.set_ref_data(chain_info, lib_info)

        key_for_signing = self.chain.get_required_keys(trx, self.wallet.get_public_keys())

        signed_transaction = self.wallet.sign_transaction(trx, key_for_signing['required_keys'], chain_info['chain_id'])
        return self.chain.push_transaction(signed_transaction)


class APIBase:

    def __init__(self, headers, base_url):
        """Base API class"""
        self.base_headers = headers
        self.base_url = base_url
        self.params = {}
        self.logger = logging.getLogger(__name__)

    def post(self, endpoint, data=None, params={}, headers={}, json={}):
        self.logger.debug(json)
        m_headers = {**self.base_headers, **headers}
        m_params = {**self.params, **params}
        return requests.post(url=parse.urljoin(self.base_url, endpoint), data=data, headers=m_headers, params=m_params,
                             json=json)


class WalletAPI(APIBase):

    def __init__(self, base_url, version):
        """Wallet API for communicating with EOSIO wallet RPCs."""
        headers = {'Accept': 'application/json', 'Content-Type': 'application/json'}
        APIBase.__init__(self, headers, parse.urljoin(base_url, '/%s/wallet/' % version))

        self.logger.debug('Constructing %s', __name__)

    # TODO: to improve exception handling, it might be best to create an error handler function
    # that throws more specific API exceptions depending on the API
    def post(self, endpoint, data=None, params={}, headers={}, json={}):
        response = super(WalletAPI, self).post(endpoint, data, params, headers, json)
        if response.status_code >= 400:
            raise WalletAPIException(response=response)
        return response.json()

    def create(self, name):
        return self.post('create', data="\"%s\"" % name)

    def unlock(self, password, name='default'):
        body = [name, password]
        return self.post('unlock', json=body)

    def lock(self, name='default'):
        return self.post('lock', data="\"%s\"" % name)

    def list_wallets(self):
        return self.post('list_wallets')

    def list_keys(self, password, name='default'):
        body = [name, password]
        return self.post('list_keys', json=body)

    def get_public_keys(self):
        return self.post('get_public_keys')

    def create_key(self, key_type="K1", name="default"):
        body = [name, key_type]
        return self.post('create_key', json=body)

    def import_key(self, private_key, name='default'):
        body = [name, private_key]
        return self.post('import_key', json=body)

    def set_timeout(self, time_out):
        return self.post('set_timeout', data=str(time_out))

    def sign_transaction(self, transaction, keys, chain_id=""):
        body = [to_dict(transaction), keys, chain_id]
        return self.post('sign_transaction', json=body)


class ChainAPI(APIBase):

    def __init__(self, base_url, version):
        """Chain API class for communicating with EOSIO based chain RPCs."""
        headers = {'Accept': 'application/json', 'Content-Type': 'application/json'}
        APIBase.__init__(self, headers, parse.urljoin(base_url, '/%s/chain/' % version))

    # TODO: to improve exception handling, it might be best to create an error handler function
    # that throws more specific API exceptions depending on the API
    def post(self, endpoint, data=None, params={}, headers={}, json={}):
        response = super(ChainAPI, self).post(endpoint, data, params, headers, json)
        if response.status_code >= 400:
            raise ChainAPIException(response=response)
        return response.json()

    def get_currency_balance(self, account, code='eosio.token', symbol='TLOS'):
        body = {'code': code, 'account': account, 'symbol': symbol}
        return self.post('get_currency_balance', json=body)

    def get_currency_stats(self, account, symbol='TLOS'):
        body = {'code': 'eosio.token', 'account': account, 'symbol': symbol}
        return self.post('get_currency_stats', json=body)

    def get_block_header_state(self, block_num):
        body = {'block_num_or_id': block_num}
        return self.post('get_block_header_state', json=body)

    def get_info(self):
        return self.post('get_info')

    def get_block(self, block_num_or_id):
        body = {'block_num_or_id': block_num_or_id}
        return self.post('get_block', json=body)

    def get_abi(self, account_name):
        body = {'account_name': account_name}
        return self.post('get_abi', json=body)

    def set_abi(self, account_name, abi_bin):
        body = {'account': account_name, 'abi': abi_bin}
        return self.post('set_abi', json=body)

    def get_code(self, account_name):
        body = {'account_name': account_name}
        return self.post('get_code', json=body)

    def set_code(self, account, code, vmtype=0, vmversion=0):
        body = {'account': account, 'vmtype': vmtype, 'vmversion': vmversion, 'code': code}
        return self.post('set_code', json=body)

    def abi_json_to_bin(self, action):
        return self.post('abi_json_to_bin', json=action)

    def abi_bin_to_json(self, bin_data):
        return self.post('abi_bin_to_json', json=bin_data)

    def get_raw_code_and_abi(self, account_name):
        body = {'account_name': account_name}
        return self.post('get_raw_code_and_abi', json=body)

    def get_account(self, account_name):
        body = {'account_name': account_name}
        return self.post('get_account', json=body)

    def get_table_rows(self, code, scope, table, output_json=True, limit=1000, lower_bound=0, upper_bound=-1):
        body = {'code': code, 'scope': scope, 'table': table, 'json': output_json, 'limit': limit,
                'lower_bound': lower_bound,
                'upper_bound': upper_bound}
        return self.post('get_table_rows', json=body)

    def get_producers(self, limit=1000, lower_bound='', output_json=True):
        body = {'limit': limit, 'lower_bound': lower_bound, 'json': output_json}
        return self.post('get_producers', json=body)

    def get_required_keys(self, transaction, available_keys):
        body = {'transaction': to_dict(transaction), 'available_keys': available_keys}
        return self.post('get_required_keys', json=body)

    def push_transaction(self, transaction):
        body = {'transaction': transaction, 'signatures': transaction['signatures'], 'compression': 'none'}
        return self.post('push_transaction', json=body)
